var Structr = require('structr'),
Mongo = require('sk/node/db/mongo').Mongo,
utils = require('sk/node/utils'),
fs = require('fs'),
TmpCache = require('sk/core/smart').smart.Cacher;



CacheManager = Structr.fh(function()
{
	var cachers = {};
	return {
		getCacher: function(type, tableName)
		{
			type = 'database';
			if(!tableName) tableName = 'default';
			
			var clazz = CacherClasses[type];
			
			if(!cachers[type]) cachers[type] = {};
			
			return cachers[type][tableName] || (cachers[type][tableName] = new clazz(tableName));
		}
	}
});

var Cacher = Structr({
	__construct: function(tableName)
	{
		this._tableName = tableName;
	},
	'final remove': function (keyOrRegexp, callback)
	{
		if(!callback) callback = function() {};
		
		if(this.testSearch(keyOrRegexp))
		{
			this.findAndRemove(keyOrRegexp, callback);
		}
		else
		{
			this.removeKey(keyOrRegexp, callback);
		}
	},
	'final get': function (keyOrRegexp, callback) 
	{ 
		if(this.testSearch(keyOrRegexp))
		{
			this.find(keyOrRegexp, callback);
		}
		else
		{
			this.getByKey(keyOrRegexp, callback);
		}
	},
	'testSearch': function(key)
	{
		return key instanceof RegExp;
	},
	'abstract find': function(regexp, callback) { },
	'abstract getByKey': function(key, callback) { },
	'abstract store': function (key, value, callback) { },
	'abstract removeKey': function(key, callback){ },
//	'abstract removeAll': function(callback){ },
	'abstract findAndRemove': function(regexp, callback) { }
});



var cdb = new Mongo('cache2');

var DBCacher = Cacher.extend({
	'override __construct': function(tableName, db)
	{
		this._super(tableName);
		
		this._collection = (db || cdb).collection('cache.' + tableName);
	},
	store: function(key, value, callback)
	{		
		this._collection.insert({ _id: key, data: value },  callback || function(){});
	},
	removeKey: function(key, callback)
	{
		var ops = {};
		
		if(key instanceof RegExp)
		{
			ops.multi = true;
		}
		
		this._collection.remove({ _id: key }, ops, callback);
	},
	findAndRemove: function(keyOrRegexp, callback)
	{
		this.removeKey(keyOrRegexp, callback);
	},
	removeAll: function(callback)
	{
		this._collection.remove(callback);
	},
	find: function(key, callback) 
	{
		this.getByKey(key, callback);
	},
	getByKey: function(key, callback)
	{
		this._collection.findOne( { _id: key }, {fields:['data']}, function(err, data)
		{
			callback(data ? data.data : null);
		});
	}
});


var MemCacher = Cacher.extend({
	'override __construct': function()
	{
		this._super();
		this._table = {};
	},
	store: function(key, value, callback)
	{
		this._table[key] = value;
		
		callback();
	},
	removeKey: function(key, callback)
	{
		delete this._table[key];
		callback();
	},
	find: function(key, callback)
	{
		for(var property in this._table)
		{
			if(property.match(regexp))
			{ 
				return callback(this._table[property], property);
			}
		}
	},
	findAndRemove: function(regexp, callback)
	{
		
		this.find(regexp, function(value, key)
		{
			delete this._table[property];
		});	
	
		callback();
		
	},
	removeAll: function(callback)
	{
		this._table = {};
		callback();
	},
	getByKey: function(key, callback)
	{
		callback(this._table[key])
	}
});


var md5NameCache = new TmpCache(1000000)

var FSCacher = Cacher.extend({
	'override __construct': function(tableName, db)
	{
		this._super(tableName);
		
		this._dir = '/tmp/cache/'+tableName;
		
		//needed for finding the write files when removing cache
		this._fnCollection = db.collection('fsCache.'+tableName);
		
		utils.mkdir_r(this._dir);
		
		//cached md5 hashes
		this._fileNames = {};
	},
	store: function(key, value, callback)
	{
		var filePath = this.getFilePath(key),
			self = this;
		fs.writeFile(filePath, value, function(err, data)
		{
			if(callback) callback();
			
			if(!err)
			{
				self._fnCollection.insert({ _id: key, filePath: filePath }, function () {});
			}
		});
	},
	removeKey: function(key, callback)
	{
		try
		{
			fs.unlink(this.getFilePath(key), callback);
			this._fnCollection.remove({ _id: key }, function(){});
		} 
		catch(e)
		{
			
		}
	},
	findAndRemove: function(regexp, callback)
	{
		var self = this;
		
		this._fnCollection.find({ _id: regexp }, function(err, cursor)
		{
			function nextFile(err, item)
			{
				if(!item)
				{
					if(!callback) callback = function() {};
					
					self._fnCollection.remove({ _id: regexp }, function(){});
					return callback();
				}
				
				try
				{
					fs.unlink(item.filePath, function()
					{
						cursor.nextObject(nextFile);
					})
				}catch(e)
				{
					cursor.nextObject(nextFile);
				}
			}
			
			cursor.nextObject(nextFile);
		});
	},
	find: function(key, callback)
	{
		var self = this;
		
		this._fnCollection.findOne({_id: regexp }, function(err, item)
		{
			if(item) return self.getByKey(item._id, callback);
			
			callback();
		});
	},
	getByKey: function(key, callback)
	{
		fs.readFile(this.getFilePath(key), 'utf8', function(err, data)
		{
			callback(data);
		});
	},
	getFilePath: function(key)
	{
		return md5NameCache.get(key) || md5NameCache.set(key,this._dir + '/' + Hash.md5(key) + '.cache', 3600*1000);
	}
})


var CacherClasses = {
	database: DBCacher,
	memory: MemCacher
};

exports.DBCacher = DBCacher;
exports.FSCacher = FSCacher;
exports.getCacher = CacheManager.getCacher;